@@ -128,7 +128,7 @@ void secure_wipe_memory(void *v, size_t n) {
 void secure_wipe_memory(void *v, size_t n) {
     memset_s(v, n, 0, n);
 }
-#elif defined(HAVE_EXPLICIT_BZERO)
+#elif HAVE_EXPLICIT_BZERO
 void secure_wipe_memory(void *v, size_t n) {
     explicit_bzero(v, n);
 }
@@ -279,7 +279,6 @@ static void *fill_segment_thr(void *thread_data)
 {
     argon2_thread_data *my_data = thread_data;
     fill_segment(my_data->instance_ptr, my_data->pos);
-    argon2_thread_exit();
     return 0;
 }
 
@@ -357,12 +356,9 @@ static int fill_memory_blocks_mt(argon2_instance_t *instance) {
     }
 
 fail:
-    if (thread != NULL) {
-        free(thread);
-    }
-    if (thr_data != NULL) {
-        free(thr_data);
-    }
+    free(thread);
+    free(thr_data);
+
     return rc;
 }
 
